@@ -126,7 +126,7 @@
 
 #define QT_FEATURE_glibc_fortify_source 1
 
-#define QT_FEATURE_trivial_auto_var_init_pattern -1
+#define QT_FEATURE_trivial_auto_var_init_pattern 1
 
 #define QT_FEATURE_stack_protector 1
 
